#ifndef SERVO_H
#define SERVO_H

#include <alternate-gpio.hpp>
#include <gpio-port.hpp>

#include "pwm-timer.hpp"
#include "servo-init.hpp"

using namespace stm32f072xb;

// usage sample
/*!
    \code {.cpp}
    ServoInit servoInit = {TIM2, OutputCompareChannel::OCC2, GpioPin::P1, AlternateFunction::AF2};
    // For the HS-422
    Servo<Port::A> servo(servoInit, 90, -90, 90, 450, 2300);
    servo.setAngle(0);
    \endcode
*/
/**
 * We found the maximum and minimum frequency range in microseconds for the following servos.
 * This is the template for displaying the informations: Servo Name : minimum pulse width-maximum pulse width
 * HS-422 : 450-2300
 * HS-485HB : 900-2100
 * MG-996R : 544-2400
 **/
template <Port pwmPort>
class Servo {
   private:
	static const uint16_t PWM_RESOLUTION = 20000;
	static const uint16_t PWM_FREQUENCY = 50;
	uint16_t _zeroDegreeDutyCycle;
	int16_t _currentAngle;
	int16_t _minAngle;
	int16_t _maxAngle;
	float _scaler;

	PwmTimer _timer;
	AlternateGpio<pwmPort> _pwmOutputPin;

   public:
	Servo(ServoInit& servoInit,
	      uint16_t amplitude,
	      int16_t minAngle,
	      int16_t maxAngle,
	      uint16_t pulseWidthMIN = 900,
	      uint16_t pulseWidthMAX = 2100)
	    : _zeroDegreeDutyCycle((pulseWidthMIN + pulseWidthMAX) / 2),
	      _currentAngle(minAngle),
	      _minAngle(minAngle),
	      _maxAngle(maxAngle),
	      _scaler((float)(_zeroDegreeDutyCycle - pulseWidthMIN) / (float)amplitude),
	      _timer(servoInit.timer, servoInit.OCC, PWM_RESOLUTION, PWM_FREQUENCY, _zeroDegreeDutyCycle),
	      _pwmOutputPin({servoInit.pwmPin, Pupdr::NONE, servoInit.alternateFunction}) {
		setAngle(_currentAngle);
	}

	void setAngle(int16_t angle) {
		if (angle > _maxAngle) {
			angle = _maxAngle;
		}
		if (angle < _minAngle) {
			angle = _minAngle;
		}
		uint16_t dutyCycle = _zeroDegreeDutyCycle + _scaler * angle;
		_timer.setDutyCycle(dutyCycle);
	}

	void incrementAngle() {
		if (_currentAngle < _maxAngle) {
			setAngle(++_currentAngle);
		}
	}
	void decrementAngle() {
		if (_currentAngle > _minAngle) {
			setAngle(--_currentAngle);
		}
	}
};

#endif
